# -*- perl -*-
#
# $Id: test_routines.pl,v 1.2 2004/01/19 17:07:28 ejb Exp $
# $Source: /home/ejb/Q/pstt/CVSroot/test-coverage/example/test_support/test_routines.pl,v $
# $Author: ejb $
#

require 5.002;
use strict;
use Cwd;
use Carp;

# TEST_VERBOSE causes failure information to go to the screen as well
# as the test log.
my $verbose = $ENV{'TEST_VERBOSE'} || 0;

# TEST_MARGIN sets the number of spaces to after PASSED or FAILED and
# before the rightmost column of the screen.
my $margin = $ENV{'TEST_MARGIN'} || 8;

my $ncols = 80;
my $got_size = 0;
eval
{
    require Term::ReadKey;
    ($ncols, undef, undef, undef) = Term::ReadKey::GetTerminalSize();
    $got_size = 1;
};
if (! $got_size)
{
    eval
    {
	# Get screen columns if possible
	no strict;
	local $^W = 0;
	local *X;
	require 'sys/ioctl.ph';
	if ((defined &TIOCGWINSZ) && open(X, "+</dev/tty"))
	{
	    my $winsize = "";
	    if (ioctl(X, &TIOCGWINSZ, $winsize))
	    {
		(undef, $ncols) = unpack('S4', $winsize);
		$got_size = 1;
	    }
	    close(X);
	}
    };
}

my $color_reset = "";
my $color_green = "";
my $color_yellow = "";
my $color_red = "";

eval
{
    require Term::ANSIColor;
    $color_reset = Term::ANSIColor::RESET();
    $color_green = Term::ANSIColor::GREEN();
    $color_yellow = Term::ANSIColor::YELLOW();
    $color_red = Term::ANSIColor::RED();
};

open(TTY, "+</dev/tty") || die "Can't open /dev/tty\n";
select((select(TTY), $| = 1)[0]);

$SIG{"INT"} = \&quit;

my $testdir;
my $tmpfile;
my $category = "";
my $tests;

my ($passes, $fails, $xpasses, $xfails) = (0, 0, 0, 0);
my $pass = 1;
my $fail = 2;
sub pass { $pass; };
sub fail { $fail; };
my $testnum = 1;

&open_log();

#
# We'll be running lots of tests programs and looking at their output.
# Make sure output is unbuffered.
#
$| = 1;

#
# Usage: runtest command exp_output exp_status exp_outcome
# Runs command and expects output exp_output and return code
# exp_status.  exp_status may be undef to indicate that status is
# ignored.
#
# If both match, the test passes; otherwise the test fails.
# exp_outcome should be either $pass or $fail.  If the test passes and
# is expected to pass, this is a pass.  If the test passes and is expected
# to fail, this is an unexpected pass.  If the test fails and is expected
# to pass, this is a fail.  If the test fails and is expected to fail,
# this is an expected fail.  Fails and unexpected passes are considered
# error conditions.
#
sub runtest
{
    return 0 unless -d $testdir;

    if (! ((@_ == 5) &&
	   ((! defined $_[3]) || ($_[3] =~ m/^\d+$/))))
    {
	print TTY "skipping: invalid arguments to runtest\n";
	return;
    }

    my($name, $cmd, $exp_output, $exp_status, $exp_outcome) = @_;

    &print_testid($name);
    &check_exp_outcome($exp_outcome) || return;

    my $exact = 0;
    my $exp_output_file = undef;
    if (ref($exp_output) eq "ARRAY")
    {
	$exact = 1;
	$exp_output_file = $exp_output->[0];
	if (-f "$tests/$exp_output_file")
	{
	    $exp_output = `cat $tests/$exp_output_file`;
	}
	else
	{
	    undef $exp_output;
	}
    }

    my $status = (system("$cmd > $tmpfile 2>&1") >> 8);
    my $output = `cat $tmpfile`;
    my $status_match = ((! defined $exp_status) || ($status == $exp_status));

    my $output_match;
    if ($exact)
    {
	$output_match = ((defined $exp_output) && ($exp_output eq $output));
	$exp_output = '' unless defined $exp_output;
    }
    else
    {
	# Modify output and expected output so that the arrangement of
	# whitespace is not important.
	my $mod_output = $output;
	my $mod_exp_output = $exp_output;
	$mod_output =~ s/[\s\n]+/ /sg;
	$mod_exp_output =~ s/[\s\n]+/ /sg;
	$mod_output =~ s/\s+$//;
	$mod_exp_output =~ s/\s+$//;
	$mod_exp_output = '^' . $mod_exp_output . "\$";

	$output_match = (($mod_exp_output eq "") && ($mod_output eq "")) ? 1 :
	    ($mod_output =~ m|$mod_exp_output|);
    }

    my $outcome = ($output_match && $status_match) ? $pass : $fail;
    my $okay = &update_counters($outcome, $exp_outcome);

    if (($outcome == $fail) && ($outcome != $exp_outcome))
    {
	my $result = "cwd: " . getcwd() . "\n";
	$result .= "command: $cmd\n";
	if (defined $exp_output_file)
	{
	    $result .= "expected output in $exp_output_file\n";
	}
	$result .= Carp::longmess();
	if (! $status_match)
	{
	    $result .= sprintf("\tExpected status: %d\n", $exp_status);
	    $result .= sprintf("\tActual   status: %d\n", $status);
	}
	if (! $output_match)
	{
	    $result .= +("--> BEGIN EXPECTED OUTPUT <--\n" .
			 $exp_output .
			 "--> END EXPECTED OUTPUT <--\n");
	    $result .= +("--> BEGIN ACTUAL OUTPUT <--\n" .
			 $output .
			 "--> END ACTUAL OUTPUT <--\n");
	    my $diff_a = "$testdir/test.exp";
	    &safe_dump($diff_a, $exp_output);
	    my $diff_b = "$testdir/test.act";
	    &safe_dump($diff_b, $output);
	    my $diff = `diff -a -u $diff_b $diff_a`;
	    $result .= +("--> DIFFERENCES FROM ACTUAL TO EXPECTED <--\n" .
			 $diff .
			 "--> END DIFFERENCES <--\n");
	}
	printf TESTLOG ("$category test %d (%s) FAILED\n%s",
			$testnum-1, $name, $result);
	if ($verbose)
	{
	    print TTY $result;
	}
    }

    $okay;
}

# Like runtest, but checks a boolean condition instead of running
# a command.
sub check_condition
{
    my($name, $condition, $exp_value, $exp_outcome) = @_;

    &print_testid($name);
    &check_exp_outcome($exp_outcome) || return;

    my $value = &$condition;
    my $outcome = ($value == $exp_value) ? $pass : $fail;
    my $okay = &update_counters($outcome, $exp_outcome);

    if (($outcome == $fail) && ($outcome != $exp_outcome))
    {
	my $result = "Condition: $condition\n";
	$result .= sprintf("Expected value: %d; actual value: %d\n",
			   $exp_value, $value);
	print TTY $result;
	printf TESTLOG ("$category test %d ($name) FAILED\n$result",
			$testnum-1);
    }

    $okay;
}

sub print_and_pad
{
    my $str = shift;
    my $spaces = $ncols - 10 - length($str) - $margin;
    $spaces = 0 if $spaces < 0;
    print TTY $str . (' ' x $spaces) . ' ... ';
}

sub print_testid
{
    my ($name) = @_;
    print_and_pad(sprintf("$category %2d (%s)", $testnum, $name));
    my $tc_filename = $ENV{'TC_FILENAME'} || "";
    if ($tc_filename && open(F, ">>$tc_filename"))
    {
	printf F "# $category %2d (%s)\n", $testnum, $name;
	close(F);
    }
    $testnum++;
}

sub check_exp_outcome
{
    my ($exp_outcome) = @_;

    if (! (($exp_outcome == $pass) || ($exp_outcome == $fail)))
    {
	printf TTY ("skipping: invalid expected outcome (%s)\n", $exp_outcome);
	0;
    }
    else
    {
	1;
    }
}

sub print_results
{
    my ($outcome, $exp_outcome) = @_;

    my $color = "";
    my $message;
    if ($outcome == $exp_outcome)
    {
	if ($outcome == $pass)
	{
	    $color = $color_green;
	    $message = "PASSED";
	}
	else
	{
	    $color = $color_yellow;
	    # " (exp)" is fewer characters than the default margin
	    # which keeps this from wrapping lines with default
	    # settings.
	    $message = "FAILED (exp)";
	}
    }
    else
    {
	if ($outcome == $pass)
	{
	    $color = $color_yellow;
	    $message = "PASSED -- fail expected";
	}
	else
	{
	    $color = $color_red;
	    $message = "FAILED";
	}
    }

    print TTY $color, $message, $color_reset, "\n";
}

sub update_counters
{
    my ($outcome, $exp_outcome) = @_;
    print_results($outcome, $exp_outcome);

    (($outcome == $pass) && ($exp_outcome == $pass)) && $passes++;
    (($outcome == $pass) && ($exp_outcome == $fail)) && $xpasses++;
    (($outcome == $fail) && ($exp_outcome == $pass)) && $fails++;
    (($outcome == $fail) && ($exp_outcome == $fail)) && $xfails++;

    ($outcome == $pass);
}

sub report
{
    if (@_)
    {
	my $expected = shift;
	print "$expected ";
    }
    print "$passes $fails $xpasses $xfails\n";
    exit 0;
}

sub open_log
{
    my $testlog = "test.log";
    open(TESTLOG, ">>$testlog") || open(TESTLOG, ">>/dev/null");
    select((select(TESTLOG), $| = 1)[0]);
}

sub print_log
{
    open_log();
    print @_;
    print TESTLOG @_;
    close(TESTLOG);
}

sub set_tmpfile
{
    $testdir = shift;
    $tmpfile = "$testdir/testout.tmp";
}

sub set_category
{
    $category = shift;
}

sub set_tests
{
    $tests = shift;
}

sub safe_dump
{
    my ($file, $output) = @_;
    open(DUMP_FILE, ">$file") or die ("can't open $file for writing");
    print DUMP_FILE ($output) or die ("can't write in $file");
    close(DUMP_FILE) or die ("can't close $file");
}

sub archconfig_require
{
    my $feature = shift;
    chop(my $result = `$ENV{'MK'}/bin/archconfig-check $feature`);
    if ($? != 0)
    {
	die "test_routines.pl/archconfig_require: archconfig-check failed\n";
    }
    if ($result eq '0')
    {
	print TTY "archconfig feature $feature missing; skipping tests\n";
	report(0);
    }
    elsif ($result eq '1')
    {
	return;
    }
    else
    {
	die "archconfig_require: " .
	    "archconfig-check returned unknown result $result\n";
    }
}

sub thread_cleanup
{
    system("rm -f *.thread-out *.seq-out *.thread-in *.seq-in");
}

sub thread_test
{
    # This function generates n + 2 tests, where n is the number of
    # threads and sequences combined.

    my ($outfile, $combined_in, $ato_out) = @_;
    my $analyze = $ENV{'MK'} . "/bin/analyze_threaded_output";
    my $filter = $ENV{'MK'} . "/bin/filter_threaded_output";
    require File::Copy;
    thread_cleanup();
    runtest("analyze $outfile with $combined_in",
	    "$analyze -split $combined_in $outfile",
	    ["$ato_out"], 0, &pass);

    my @files = ();
    if (open(A, "<$ato_out"))
    {
	my $suf = "";
	while (<A>)
	{
	    if (m/^threads:/)
	    {
		$suf = "thread";
	    }
	    elsif (m/^sequence /)
	    {
		$suf = "seq";
	    }
	    elsif (m/^\s+(\S+)/)
	    {
		push(@files, "$1.$suf");
	    }
	}
	close(A);
    }

    my $okay = 1;

    foreach my $f (@files)
    {
	my $in = "$f-in";
	my $out = "$f-out";
	if (! runtest("check $f",
		      "cat $out",
		      ["$in"], 0, &pass))
	{
	    $okay = 0;
	}
    }

    if ($okay && (! exists $ENV{'DUMP_THREADED_OUTPUT'}))
    {
	runtest("all files okay",
		"echo yes",
		"yes\n", 0, &pass);
    }
    else
    {
	# This gives us a handy way of getting the actual program
	# output into the test log and comparing it with the combined
	# input.  This test case will always fail but can still
	# provide useful information.

	my $tmp = "__tmp__.test-thread";
	system("$filter $combined_in > $tmp");
	runtest("program output",
		"cat $outfile",
		[$tmp], 0, &pass);
	unlink $tmp;
    }
}

1;
